loss_id( loss_01, id ).    				% identity
loss_id( loss_eg_01, ed ). 				% equivalence class identity
loss_id( loss_ck, ck ).    				% Castelo and Ko\v{c}ka
loss_id( loss_ks, ks ).					% Koivisto and Sood
loss_id( loss_naive_edgecount, ne ). 
loss_id( loss_kl, kl ).					% KL divergence

:- ensure_loaded(library(lists)).
:- ensure_loaded(library(ugraphs)).
:- ensure_loaded(library(ordsets)).

/* We assume num_rvs/1 is present
:- dynamic num_rvs/1.

:-
	data_format(TermFormat),
	functor(TermFormat,_,Arity),
	assert(num_rvs(Arity)).
	*/
	

/*---------------------------------------------------------------------

ALL CHAIN GRAPHS (INCLUDING BNS) ARE ORDERED SETS OF EDGES WHERE
A-B INDICATES AN ARROW FROM A TO B

AN UNDIRECTED EDGE BETWEEN A AND B IS GIVEN BY TWO TERMS: A-B AND B-A

James Cussens
Sat Jul 24 10:19:03 BST 2004

true_eg/1 is not defined in this file
true_bn/1 is not defined in this file
---------------------------------------------------------------------*/

/*---------------------------------------------------------------------
  Predicate: loss_01(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: 'straw man' loss, only the true BN has zero loss, all others
             have loss one.
---------------------------------------------------------------------*/
% pre_loss_01( BN, BN ).

loss_01(BaseBN,BaseBN,0) :-
	% true_bn(TrueBN),
	!.
loss_01(_BaseBN,_RelBN,1).


/*---------------------------------------------------------------------
  Predicate: loss_eg_01(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: slightly less weird than loss_01/2. For this one you have
             to be in the right Markov equivalence class to get zero
	     loss, otherwise get a loss of 1.
---------------------------------------------------------------------*/
pre_loss_eg_01( BN, EG ) :-
	bn_eg( BN, EG ).

loss_eg_01(BaseEG,BaseEG,0) :-
	% true_eg(TrueEG),
	% bn_eg(BN,TrueEG),
	!.
loss_eg_01(_BEG,_REG,1).


/*---------------------------------------------------------------------
  Predicate: loss_ck(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: Assumes that the essential graph of the true BN is available
             via true_eg
	     follows Castelo and Ko\v{c}ka
	     Perhaps we should normalise by dividing by the square of
	     the number of vertices in BN??
---------------------------------------------------------------------*/
pre_loss_ck( BN, EG ) :-
	bn_eg( BN, EG ),
	write( bn(BN) ), nl,
	write( eg(EG) ), nl.

loss_ck(BaseEG,RelEG,Loss) :-
	% true_eg(TrueEG),    % essential graph of true BN
	% bn_eg(BN,EG),
	cg_diffs(RelEG,BaseEG,0,Loss).

/*---------------------------------------------------------------------
  Predicate: loss_ks(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: Assumes that the true BN is available
             via true_bn
	     follows Koivisto and Sood
---------------------------------------------------------------------*/
% pre_loss_ks(BN,BN).

loss_ks(BaseBN,RelBN,Loss) :-
	% true_bn(TrueBN),    % true BN
	edge_count(ks,BaseBN,RelBN,NumExtra,NumMissing),
	AbsLoss is NumExtra + NumMissing,
	num_rvs(N),
	Loss is AbsLoss/(N**2).


/*---------------------------------------------------------------------
  Predicate: loss_naive_edgecount(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: Assumes that the true BN is available
             via true_bn
	     like loss_ks/2 but doesn't bother to eliminate covered
	     edges
---------------------------------------------------------------------*/
% pre_loss_naive_edgecount(BN,BN) :-

loss_naive_edgecount(BaseBN,RelBN,Loss) :-
	% true_bn(TrueBN),    % true BN
	edge_count(naive,BaseBN,RelBN,NumExtra,NumMissing),
	AbsLoss is NumExtra + NumMissing,
	num_rvs(N),
	Loss is AbsLoss/(N**2).

/*---------------------------------------------------------------------
  Predicate: loss_kl(+BN,-Loss)
       Args: BN is a BN (a DAG)
             -Loss is the loss suffered by taking BN to be the true BN
   Succeeds: exactly once
   Comments: Assumes that the parameterised true BN is available
             and that data is available.
             Computes the KL divergence between the probability distribution
	     defined by the true BN, and that given by fitting parameters to
	     the input BN using maximum likelihood
	     NOT YET IMPLEMENTED!
---------------------------------------------------------------------*/
loss_kl(BN,Loss) :-
	ptrue_bn(ParameterisedTrueBN),    % true BN
	data(Data),
	ml_fit(BN,Data,ParameterisedBN),
	kl_divergence(ParameterisedTrueBN,ParameterisedBN,Loss).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%













/*---------------------------------------------------------------------
  Predicate: cg_diffs(+CG1,+CG2,+In,-Out)
       Args: CG1 and CG2 are chain graphs (eg essential graphs)
                 each represented as an ordered set of edges
             Out is a count of the number of differences
	     In is just an accumulator
   Succeeds: exactly once
   Comments: considers all pairs of vertices, if they are differently
             connected then this is a difference
	     First go through edges in CG1 counting any differences
	     for these with CG2, deleting corresponding edges in CG2
	     then count the edges left in CG2
---------------------------------------------------------------------*/
cg_diffs([],CG2,In,Out) :-
	count_edges(CG2,In,Out).
cg_diffs([A-B|Rest],CG2,In,Out) :-
	ord_try_del_element(Rest,B-A,Left,Result),  %directed or undirected?
	ord_try_del_element(CG2, A-B,CG21,ResultAB),
	ord_try_del_element(CG21,B-A,CG22,ResultBA),
	check_discrepancy(Result,ResultAB,ResultBA,Discrepancy),
	Mid is In + Discrepancy,
	cg_diffs(Left,CG22,Mid,Out).

/*---------------------------------------------------------------------
  Predicate: check_discrepancy(+Result,+ResultAB,+ResultBA,-Discrepancy)
       Args: 
   Succeeds: exactly once
   Comments: 
---------------------------------------------------------------------*/
check_discrepancy(yes,yes,yes,0) :- !. % undirected in both
check_discrepancy(no, yes,no, 0) :- !. % directed   in both
check_discrepancy(_ , _  ,_ , 1).      % else, a discrepancy

/*---------------------------------------------------------------------
  Predicate: count_edges(CG,+In,-Out)
       Args: 
   Succeeds: exactly once
   Comments: Count edges in CG without double counting undirected ones
---------------------------------------------------------------------*/
count_edges([],Out,Out).
count_edges([A-B|Rest],In,Out) :-
	ord_try_del_element(Rest,B-A,Left,_Result),  %don't double count
	Mid is In + 1,
	count_edges(Left,Mid,Out).
	  

/*---------------------------------------------------------------------
  Predicate: bn_eg(+BN,-EssentialGraph)
       Args: BN is a BN (a DAG)
             -Essential graph for BN
   Succeeds: exactly once
   Comments: Algorithm from Andersson et al
             All graphs represented as ordered set of edges
	     Undirected edge is just two directed edges
---------------------------------------------------------------------*/
bn_eg(CG,EG) :-
	find_linemakers(CG,LineMakers),
	(
	  LineMakers = []	%no further progess
	->
	  EG = CG
	;
	  ord_union(CG,LineMakers,CG1),
	  bn_eg(CG1,EG)
	).


/*---------------------------------------------------------------------
  Predicate: find_linemakers(+CG,-LineMakers)
       Args: 
   Succeeds: exactly once
   Comments: find arrows B-A which can be 'added to other arrows' A-B to
             make undirected edges, because A-B is not strongly protected
---------------------------------------------------------------------*/
find_linemakers(CG,LineMakers) :-
	findall(B-A,
		(
		  member(A-B,CG),
		  \+ ord_member(B-A,CG),
		  \+ strongly_protected(A-B,CG)
		),
		LineMakers1
	       ),
	list_to_ord_set(LineMakers1,LineMakers).
	
/*---------------------------------------------------------------------
  Predicate: strongly_protected(+Arrow,+CG)
       Args: 
   Succeeds: 0 +
   Comments: succeeds if Arrrow is strongly protected in CG
             only correct of Arrow is an arrow
---------------------------------------------------------------------*/
strongly_protected(A-B,CG) :-
	member(C-B,CG),         % case (b) and (c) in Andersson et al
	\+ C = A,
	\+ ord_member(B-C,CG),
	\+ ord_member(C-A,CG).
strongly_protected(A-B,CG) :-
	member(C-A,CG),         % case (a) in Andersson et al
	\+ ord_member(A-C,CG),
	\+ ord_member(B-C,CG),
	\+ ord_member(C-B,CG).
strongly_protected(A-B,CG) :-
	member(C1-B,CG),        % case (d) in Andersson et al
	\+ ord_member(B-C1,CG),
	member(C2-B,CG),
	\+ C1 = C2,
	\+ ord_member(B-C2,CG),
	ord_member(A-C1,CG),
	ord_member(C1-A,CG),
	ord_member(A-C2,CG),
	ord_member(C2-A,CG).

/*---------------------------------------------------------------------
  Predicate: edge_count(+Type,+BNTrue,+BN,-NumExtra,-NumMissing)
       Args: Type is naive or ks; naive just counts all differences, but
             ks filters out "covered" (see Chickering) edges from the count
             BNTrue and BN are ordered sets of edges
             -Extra   counts no. of edges in BN not in BNTrue
	     -Missing counts no. of edges in BNTRue not in BN
   Succeeds: exactly once
   Comments: Replicates Koivisto and Sood
             BNTrue = A -> B -> C 
	     BN =     A <- B <- C
	     has one extra and one missing edge according to this,
	     we can reverse B <- C, but not
	     A <- B
	     This is for Koivisto and Sood
---------------------------------------------------------------------*/
edge_count(Type,BNTrue,BN,NumExtra,NumMissing) :-
	ord_subtract(BN, BNTrue, Extra  ),    %library predicate
	ord_subtract(BNTrue, BN, Missing),
	(
	  Type = naive
	->
	  RealExtra = Extra,
	  RealMissing = Missing
	;
	  filter_covered(Extra,BN,Missing,RealExtra,RealMissing)
	),
	length(RealExtra,NumExtra),
	length(RealMissing,NumMissing).

/*---------------------------------------------------------------------
  Predicate: filter_covered(+Extra,+BN,+Missing,-RealExtra,-RealMissing)
       Args: 
   Succeeds: exactly once
   Comments: covered arrows are removed from Extra and their reverses
             are removed from Missing
---------------------------------------------------------------------*/
filter_covered([],_BN,Missing,[],Missing).
filter_covered([A-B|Rest],BN,Missing,RealExtra,RealMissing) :-
	ord_del_element(Missing,B-A,MRest),       % 'extra' edge result of a reversal
	findall(PA,member(PA-A,BN),PAs),          % gather parents of A
	list_to_ord_set([A|PAs],PAsSet),          % make A+parents of A into a set
	findall(PB,member(PB-B,BN),PBs),          % find B's parents
	list_to_ord_set(PBs,PAsSet),              % B's parents = A+A's parents
	!,                                        % 'Extra' edge is covered
	filter_covered(Rest,BN,MRest,RealExtra,RealMissing).
filter_covered([Edge|Rest],BN,Missing,[Edge|RealExtra],RealMissing) :-
	filter_covered(Rest,BN,Missing,RealExtra,RealMissing).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% U T I L I T Y


/*---------------------------------------------------------------------
  Predicate: ord_try_del_element(+In,+Elt,-Out,-Result)
       Args: 
   Succeeds: exactly once
   Comments: Try and delete an element from an ordered set, Result
             says whether you did.
---------------------------------------------------------------------*/

ord_try_del_element(In,Elt,Out,yes) :-
	ord_del_element(In,Elt,Out),
	!.
ord_try_del_element(In,_Elt,In,no).
